;;;; Copyright (C) 1998 R. Dennis Cook and Sanford Weisberg.  Arc is 
;;;; XLISP-STAT computer code to accompany the book "An Introduction to
;;;; Regression Graphics", published by John Wiley, Inc.
;;;; All rights reserved.

;;;; $Revision: 1.3 $

;;;
;;;  Added Variable Plots
;;;

(defmeth regression-model-proto :avps (&key (name (send self :name)))
"Message args:  ()
Creates a plot with a slide-bar, giving all
added variable plots for adding a predictor last.  Does computations on the
fly, so if computing method is other than LS this can be very slow."
 (cond ((= (length (send self :basis)) 1) (send self :avp1))
       (t
  (let* ((names (send self :predictor-names))
         (basis (send self :basis))
         (dataset (send self :dataset))
         (index (first basis))
         (y-name (send self :response-name))
         (title (format nil "(~a)Added Variable Plots" name))
         x reg plot reg1 b plot
         (data #'(lambda ()
           (setf reg (send self :make-clone))
           (setf x (column-list (send self :x)))
           (setf b (if (send self :intercept) 
                       (rest (send self :coef-estimates))
                       (send self :coef-estimates)))
           (setf reg1 (regression-model (select x (remove index basis))
                                        (select x index) :print nil
                                        :included (send reg :included)))
           (send reg :x (select x (remove index basis)))
           (send reg :compute) 
           (when plot (send plot :add-slot 'slope (select b index)))
           (list (send reg1 :residuals) (send reg :residuals))))
         (plot (send dataset :make-plot data (list
           (make-res-label (select names index)
                   (select names (remove index basis)))
           (make-res-label y-name (select names (remove index basis))))
                     :weights (send self :pweights)
                     :title title))
         (slider (first (send plot :add-overlay 
                  (send slider-control-proto :new basis
                       :title ""
                       :length (send plot :slider-width)
                       :location (send plot :locate-next-control :height 2)
                       :display (select names basis))))))
  (send plot :start-next-frame #'(lambda () (send plot :start-buffering)))
  (send plot :finish-next-frame #'(lambda () (send plot :adjust-to-data)
                                          (send plot :buffer-to-screen)))
  (defmeth slider :do-action (j)
           (setf index j)
    (when (slot-value 'graph)
           (send plot :add-slot 'slope (select b index))
           (send plot :variable-label 0
             (make-res-label (select names index) 
                     (select names (remove index basis))))
           (send plot :variable-label 1
             (make-res-label y-name (select names (remove index basis))))
           (send reg :x (select x (remove index basis)))
           (send reg :compute)
           (send reg1 :x (select x (remove index basis)))
           (send reg1 :y (select x index))
           (send plot :draw-next-frame '(0 1)
                 (list (send reg1 :residuals) (send reg :residuals)))))
        plot))))

(defmeth regression-model-proto :avp1 (&key (plot-controls t))
  (let* ((data #'(lambda ()
           (let* ((inc (if-else (send self :included) 1 0))
                  (x (first (column-list (send self :x))))
                  (wt (if (send self :weights) (send self :weights)
                          (repeat 1 (length (send self :y)))))
                  (y (send self :yvar)))
                 (list (- x (/ (sum (* inc wt x)) (sum (* inc wt))))
                       (- y (/ (sum (* inc wt y)) (sum (* inc wt))))))))
         (plot (send self :make-plot data (list 
          (format nil "Centered ~a" (first (send self :predictor-names)))
          (format nil "Centered ~a" (send self :response-name)))
              :title (format nil "(~a)Added Variable Plot" (send self :name))
              :plot-controls plot-controls)))
; (when (and plot-controls (eql (send self :slot-value 'proto-name)
;           'regression-model-proto)) (send plot :avp-controls))
        plot))

(defun make-res-label (a b &key (length 45))
  (let* ((lab (format nil "e(~a|~a)" a b)))
        (if (< (length lab) length) lab
            (format nil "e(~a|others)" a))))

;;;
;;;  3D added variable plots
;;;

(defmeth regression-model-proto :avp3d (&rest args) 
"Args: (&rest args)
Produces a 3D added variables plot, with each group consisting of a single
variable.  args are passed to the plot."
    (let* ((basis (send self :basis))
           (names (select (send self :predictor-names) basis))
           add)
    (cond ((= (length names) 2)
           (apply #'send self :plot (first names) (send self :response-name)
                  (second names) args))
          (t (setf add (let ((sel (avp3d-dialog names)))
                     (if (null sel) (return-from :avp3d) sel)))
              (apply #'send self :avp3d1 add args)))))

(defun avp3d-dialog (names &key (text "add"))
  (let* ((ans (twolist-dialog names :title 
                    (format nil "Select two terms to ~a" text)
                    :right-initial (repeat "" 2))))
         (cond ((null ans) nil)
           ((or (null (second ans)) (< (length (second ans)) 2))
            (message-dialog "Two variables must be specified")
            (avp3d-dialog names :text text))
           (t ans))))

(defmeth regression-model-proto :avp3d1 (names &rest args
           &key (name (send self :name))) 
  (let* ((data (send self :data))
         (basis (send self :basis))
         (preds (select (send self :predictor-names) basis))
         (xsel (mapcar #'(lambda (a) (position a preds :test #'equal)) 
               (first names)))
         (ysel (mapcar #'(lambda (a) (position a preds :test #'equal)) 
               (second names)))
         (x (select (column-list (send self :x)) xsel))
         (y (select (column-list (send self :x)) ysel))
         (zreg (let ((r (send self :make-clone))) (send r :x x) r))
         (xreg (regression-model x (first y) :print nil))
         z1
         (data #'(lambda () 
           (send zreg :add-slot 'included (send self :included))
           (defmeth zreg :included (&rest args)
             (apply #'call-method regression-model-proto :included args))
           (send xreg :included (send self :included))
           (send xreg :y (first y)) 
           (setf z1 (send xreg :residuals)) 
           (send zreg :included (send self :included))
           (send xreg :y (second y)) 
           (list z1 (send zreg :residuals) (send xreg :residuals))))
         (labels (list (format nil "e(~a|rest)" (first (second names)))
                       (format nil "e(~a|rest)" (send self :response-name))
                       (format nil "e(~a|rest)" (second (second names)))))
         (plot (apply #'send self :make-plot data labels
                     :weights (send self :pweights)
                     :title (format nil "(~a)AVP-3D" name) args)))
        plot))



