;;; $Revision: 1.1 $


(require "glim")

;;
;;  Enhancements to glim-proto by S. Weisberg
;;

;; overrides for standard methods
(defmeth glim-proto :isnew (&rest args &key y link
                                 (offset 0)
                                 pweights
                                 (verbose t)
                                 (recycle nil))
"This override works with the override for the regression-model-proto :isnew
method, but not with the original :isnew method."
  (if link (send self :link link))
  (send self :yvar y)
  (send self :offset offset)
  (send self :pweights pweights)
  (send self :recycle recycle)
  (send self :verbose verbose)  
  (apply #'call-next-method args)
  self)

(defmeth glim-proto :x (&rest args)
  (apply #'call-next-method args))

(defmeth glim-proto :display (&rest args ) 
  (apply #'send self :display-estimates args))
(defmeth glim-proto :display-estimates ()
  (call-next-method :r-squared nil :t-value " Est/SE")
  (format t "Pearson X2:           ~10,3f~%" (send self :pearson-x2))
  (format t "Deviance:             ~10,3f~%" (send self :deviance)))
(defmeth normalreg-proto :display-estimates (&rest args)
 (cond ((eq (send self :link) identity-link)
        (apply #'call-method regression-model-proto :display-estimates args))
       (t (call-next-method))))

(defmeth normalreg-proto :display (&rest args)
 (cond ((eq (send self :link) identity-link)
        (call-method regression-model-proto :display))
       (t (call-next-method :anova nil))))

(defmeth normalreg-proto :fit-means (&rest args)
 (cond ((eq (send self :link) identity-link)
        (call-method regression-model-proto :fit-means))
       (t (apply #'call-next-method args))))

(defmeth glim-proto :response-name (&optional name)
  (cond (name (call-next-method name))
        (t (let ((name (call-next-method)))
                (if name name "Response")))))

(defmeth glim-proto :mean-deviance ()
"Message args: ()
Returns mean deviance for included cases, adjusted for degrees of
freedom, or zero for saturated models."
  (cond ((> (send self :df) 0) (/ (send self :deviance) (send self :df)))
        (t (send self :deviance))))

(defmeth glim-proto :response () (send self :yvar))

(defmeth glim-proto :plotlist () *glm-plotlist*)

(defmeth binomialreg-proto :plotlist () *binomial-plotlist*)

(defmeth normalreg-proto :plotlist ()
  (if (eq (send self :link) identity-link) 
      (call-method regression-model-proto :plotlist)
      (call-next-method)))

(defmeth normalreg-proto :yvar (&optional (new nil set))
  (cond ((eq (send self :link) identity-link) 
         (call-method regression-model-proto :y new))
        (set (call-next-method new))
        (t (call-next-method))))

(defmeth binomialreg-proto :obs-fraction ()
  (div (send self :yvar) (send self :trials)))

(defmeth binomialreg-proto :fit-fraction ()
  (div (send self :fit-means) (send self :trials)))

(defmeth glim-proto :dev-residuals ()
  (send self :deviance-residuals))

(defmeth glim-proto :deviance-residuals ()
  "Message args: ()
  Returns the components of deviance residuals for non binomial models."
    (let* ((dev (sqrt (pmax 0 (send self :deviances))))
           (sign (if-else (< (send self :yvar) (send self :fit-means)) -1 1)))
     (* sign dev)))

(defmeth glim-proto :standardized-deviance-residuals ()
"Message args: ()
Returns the standardized deviance residuals, (Davison and Tsai, 1989)."
 (flet ((/+ (a b)
            (cond ((< (abs a) machine-epsilon) 0)
                  ((< (abs b) machine-epsilon) 0)
                  (t (/ a b)))))
  (let* ((dev (send self :deviance-residuals))
         (inc (send self :included))
         (s (send self :scale))
         (h (pmin (- 1 machine-epsilon) (send self :leverages))))
    (mapcar #'(lambda (inc dev h) (if inc
             (/+ dev (sqrt (* s (- 1 h))))
             (/+ dev (sqrt (* s (+ 1 h))))))
            inc dev h))))

(defmeth normalreg-proto :g2-residuals ()
  (cond ((eq (send self :link) identity-link) 
         (call-method regression-model-proto :residuals))
        (t (send self :call-next-method))))

(defmeth glim-proto :g2-residuals ()
"Message args: ()
Returns  a weighted combination of the standardized deviance and chi
residuals, (Davison and Tsai, 1989)."
  (let* ((dev (send self :standardized-deviance-residuals))
         (chi (send self :standardized-chi-residuals))
         (inc (send self :included))
         (h (send self :leverages))
         (sign (with-missing #'(lambda (d) (if-else (< d 0) -1 1)) dev)))
    (mapcar #'(lambda (inc sign h dev chi)
     (if inc (* sign (sqrt (+ (* (- 1 h) (^ dev 2)) (* h (^ chi 2)))))
             chi))
            inc sign h dev chi)))

(defmeth glim-proto :T-residuals () (send self :g2-residuals))

(defmeth glim-proto :std-chi-res ()
  (send self :standardized-chi-residuals))

(defmeth glim-proto :pearson-x2 ()
 (sum (^ (select (send self :residuals) (which (send self :included))) 2)))

(defmeth glim-proto :mean-pearson-x2 ()
  (/ (send self :pearson-x2) (send self :df)))

(defmeth glim-proto :summary-plot ()
 (let* ((n (send self :name)) 
        (p (send (send self :data) :plot (format nil "~a:eta" n)
                                        (format nil "~a:response" n))))
   (send p :link-control self)
   p))

(defmeth binomialreg-proto :summary-plot ()
 (let* ((n (send self :name)) 
        (p (send (send self :data) :plot (format nil "~a:eta" n)
                                        (format nil "~a:obs-fraction" n))))
   (send p :link-control self)
   p))

(defmeth scatterplot-proto :link-control (reg)
"Adds a button to a plot of 'eta versus 'yvar in glim-proto."
  (let* ((lline #'(lambda (args)
                       (let* ((eta1 (select (send reg :eta)
                                      (which (send reg :included))))
                              (x (rseq (min eta1) (max eta1) 50))
                              (y (send (send reg :link) :means x)))
                             (list x y))))
         (control (send curve-control-proto :new
               (send self :locate-next-control)
               "Mean function" lline (list nil) (nth 6 *colors*))))
        (send self :add-overlay control)
        (send self :finish-next-frame #'(lambda ()
                      (when (send control :slot-value 'selected)
                            (send control :draw-curve))))
        (send self :redraw)))

(defmeth normalreg-proto :c+r-residuals ()
  (cond ((eq (send self :link) identity-link) 
         (call-method regression-model-proto :c+r-residuals))
        (t (send self :call-next-method))))

; corrected 12/13/95
(defmeth glim-proto :c+r-residuals ()
  (let ((d (if (send self :pweights)
               (* (send self :pweights) (sqrt (send self :weights)))
               (sqrt (send self :weights)))))
  (/ (send self :residuals) d)))

(defmeth normalreg-proto :ares (&rest args)
  (cond ((eq (send self :link) identity-link) 
         (apply #'call-method regression-model-proto :ares args))
        (t (apply #'send self :call-next-method args))))

(defmeth normalreg-proto :ares1 (&rest args)
  (cond ((eq (send self :link) identity-link) 
         (apply #'call-method regression-model-proto :ares1 args))
        (t (apply #'send self :call-next-method args))))

(defmeth glim-proto :ares (&rest args)
"Produces and ARES plot (Cook and Weisberg, 1994, \"ARES plots for generalized
linear models,\" Computational Statistics and Data Analysis, 17, 303-316."
    (let* ((names (send self :predictors))
           (sel (let ((s (select-list-dialog names
                          :title "ARES plot of..."
                          :left-title "Candidates"
                          :right-title "Selections...")))
                  (if (null s) (return-from :ares))
                  (if (null (second s)) (return-from :ares))
                  s))
           (base (select names (first sel)))
           (add (select names (second sel))))
          (apply  #'send self :ares1 base add args)))

(defmeth glim-proto :ares1 (base add &rest args &key 
                  (lambda (combine 0 .05 (rseq .1 .9 9) .95 1))
                  (plot-controls t) (x-axis :eta) (y-axis :g2-residuals) )
    (let* (x1 x2 y1 y2 m
           (missing (find-missing (send self :residuals)))
           (d (send self :data)) 
           (b (cond ((and (null base) (send self :intercept)) '("Ones"))
                    (t base)))
           (data #'(lambda () 
             (setf m (send self :make-clone))
             (if (null base) (send m :intercept nil))
             (send m :x (apply #'append (mapcar 
                 #'(lambda (n) (send d :data n)) b)))  
             (setf y1 (send m y-axis))
             (setf x1 (send m x-axis))
             (setf y2 (send self y-axis))
             (setf x2 (send self x-axis))
             (list x1 y1)))
           (title (format nil "ARES:~a + ~a" base add))
           (xrange #'(lambda (x) (list (min x) (max x))))
           (labels (mapcar #'(lambda (a) (methname-to-string a))
                           (list x-axis y-axis)))
           (ares-plot (apply #'send self :make-plot data labels
                    :title title :plot-controls plot-controls args))
           (slider (send ares-plot :ares-plot-control 
                         (iseq (length lambda)) lambda plot-controls)))
      (apply #'send ares-plot :nice-range 0 (funcall xrange (combine x1 x2)))
      (apply #'send ares-plot :nice-range 1 (funcall xrange (combine y1 y2)))
      (when missing (send ares-plot :point-missing missing 
                    (repeat t (length missing))))
      (send ares-plot :redraw)
      (defmeth slider :do-action (lam) 
        (let* ((l (select lambda lam))
               (x (+ (* (- 1 l) x1) (* l x2)))
               (y (+ (* (- 1 l) y1) (* l y2))))
        (send ares-plot :draw-next-frame '(0 1) (list x y))))
      ares-plot))

;;;
;;;  Functions for fitting quasi-symmetry models and
;;;  Quasi independence to square contingency tables with
;;;  the cells in row-order or column-order
;;;  Fixed, May 28, 1997
;;;
 
(defun quasi-indep (n &optional (cell-order (iseq (^ n 2))))
"Function args:    (n &optional (cell-order (iseq (^ n 2))))
Returns a list of length n^2 that is zero for all off-diagonal cells
and a different integer for each diagonal cell in a n by n square contingency
table.  Cells are assumed to be in order by rows, or in the order given by
cell-order."
  (let* ((ans (make-list (^ n 2) :initial-element 0)))
   (dolist (j (iseq n)) 
           (setf (select ans (position (* j (+ n 1)) cell-order)) (1+ j)))
   ans))

(defun quasi-symmetry (n &optional (cell-order (iseq (^ n 2))))
"Function args: (n &optional (cell-order (iseq (^ n 2))))
Returns a list of length n^2 that has
a different integer for each diagonal cell in a n by n square contingency
table, and the same integer for symmetric off-diagonal cells.  Cells are 
assumed to be in order by rows, or in the order given by cell-order."
  (let* ((ans (quasi-indep n cell-order))
         k pos)
    (dolist (i (iseq 0 (- n 2)))
     (dolist (j (iseq (+ i 1) (- n 1)))
      (setf k (format nil "~d,~d" i j))
      (setf pos (position (+ (* n i) j) cell-order))
      (setf (select ans pos) k)
      (setf pos (position (+ (* n j) i) cell-order))
      (setf (select ans pos) k)))
    ans))

;;;
;;;
;;;

(defmeth glim-proto :get-fit-variances ()
   (let* ((eta (send self :eta))
     (mu (send self :fit-means eta))
     (v (send self :fit-variances mu)))
      v))

(defmeth normalreg-proto :verbose (&rest args)
 (cond ((eq (send self :link) identity-link) nil)
       (t (apply #'call-next-method args))))

(defmeth normalreg-proto :compute ()
 (cond ((eq (send self :link) identity-link) 
        (defmeth self :weights (&rest args) 
         (apply #'send self :pweights args))
        (call-method regression-model-proto :compute)
        (send self :set-eta)
        (send self :scale (send self :fit-scale)))
       (t (defmeth self :weights (&rest args) (apply #'call-next-method args))
         (call-next-method))))

(defmeth normalreg-proto :deviance () 
   (if (eq (send self :link) identity-link) 
       (send self :residual-sum-of-squares)
       (call-next-method)))

(defmeth binomialreg-proto :compute-step ()
"Args: ()
Executes one iteratively reweighted least squares step."
  (let* ((yvar (send self :yvar))
         (offset (send self :offset))
         (eta (send self :eta))
         (trials (send self :trials))
         (eps .0000001)
         (mu (pmax (pmin (send self :fit-means eta) (- trials eps)) eps))
         (d-eta (send self :fit-link-derivs mu))
         (z (- (+ eta (* (- yvar mu) d-eta)) offset))
         (v (send self :fit-variances mu))
         (w-inv (* d-eta d-eta v))
         (pw (send self :pweights)))
    (send self :y z)
    (send self :weights (if pw (div pw w-inv) 
                               (div (repeat 1 (length w-inv)) w-inv)))
    (call-method regression-model-proto :compute)
    (send self :set-eta)
    (send self :set-deviances)
    (if (send self :estimate-scale) 
        (send self :scale (send self :fit-scale)))))

;;;
;;; Methods to fit extra-binomial variation
;;;

(defmeth binomialreg-proto :estimate-extra-variation ()
"Method args: ()
Uses Williams (1982, Applied Statistics, 144-148) quasi likelihood model to
estimate an extra variation parameter, phi."
  (let* ((pw (if (send self :pweights) (send self :pweights)
                 (repeat 1 (send self :num-cases))))
         (inc (which (send self :included)))
         (h (send self :leverages))
         (trials (send self :trials))
         (x2 (send self :pearson-x2)))
        (div (- x2 (sum (select (* pw (- 1 h)) inc)))
             (sum (select (* pw (- trials 1) (- 1 h)) inc)))))

(defmeth binomialreg-proto :extra-variation-iterate ()
 (let* ((phi (send self :estimate-extra-variation))
        (pw (div (+ 1 (* phi (- (send self :trials) 1)))))
        (data (send self :data))
        (dl (send data :find-datalist "Extra-bin-var")))
  (send self :pweights pw)
  (if dl (send dl :data pw) (send data :bin-wts pw))
  (send self :args (append (list :weights "Extra-bin-var")
        (send self :args)))
   phi))

(defmeth binomialreg-proto :fit-extra-variation (&key (maxcount 20))
 (cond ((=  1 (max (select (send self :trials) (which (send self :included)))))
        (message-dialog "Can't fit extra variation with
number of trials equal to 1")
        (return-from :fit-extra-variation))
       (t
 (let ((ans 0))
 (do ((count 1 (+ count 1))
      (phi-old -1 phi)
      (phi (send self :extra-variation-iterate)
           (send self :extra-variation-iterate)))
     ((or (> count maxcount) (< (abs (/ (- phi-old phi) phi)) .001))
      (setf ans phi)))
  (defmeth self :display (&rest args) (apply #'call-next-method args)
     (format t "Extra variation ~25t~11,6g~%" (optfmt0 ans)))
  (send self :display)))))

;;; div is / except on Windows.  For missing values.

(defmeth binomialreg-proto :initial-probs ()
  (let* ((n (send self :trials))
    (p (div (pmax (pmin (send self :yvar) (- n 0.5)) 0.5) n)))
    p))

(defmeth binomialreg-proto :fit-variances (mu)
  (let* ((n (send self :trials))
         (p (div mu n)))
    (* n p (- 1 p))))

(defmeth binomialreg-proto :fit-probabilities ()
"Message args: ()
Returns the fitted probabilities for the model."
  (div (send self :fit-means) (send self :trials)))

(defmeth binomialreg-proto :fit-prob () (send self :fit-probabilities))

;; Modified for missing data on Windows

;;;;
;;;; Logit Link Prototype
;;;;

(defmeth logit-link :eta (p) 
 (with-missing #'(lambda (p) (log (/ p (- 1 p)))) p))
(defmeth logit-link :means (eta)
 (with-missing #'(lambda (u) (/ u (1+ u))) (with-missing #'exp eta)))
(defmeth logit-link :derivs (p) (+ (div p) (div (- 1 p))))

;;;;
;;;; Probit Link Prototype
;;;;

(defmeth probit-link :eta (p) (with-missing #'normal-quant p))
(defmeth probit-link :means (eta) (with-missing #'normal-cdf eta))
(defmeth probit-link :derivs (p) (div (with-missing #'normal-dens 
                                        (with-missing #'normal-quant p))))

;;;
;;;  Complementary log-log
;;;

(defmeth cloglog-link :eta (p) 
      (with-missing #'log (- (with-missing #'log (- 1 p)))))
(defmeth cloglog-link :means (eta) 
      (- 1 (with-missing #'exp (- (with-missing #'exp eta)))))
(defmeth cloglog-link :derivs (p)
  (let ((q (- 1 p)))
    (- (div (* (with-missing #'log q) q)))))

;;;;
;;;; Compute method override--used only deviance for convergence
;;;;

(defmeth glim-proto :compute ()
  (let* ((epsilon (send self :epsilon))
         (epsilon-dev (if (boundp '*deviance-change*)
                           *deviance-change* .001))
         (min-dev (send self :min-deviance))
         (maxcount (send self :count-limit))
         (low-lim (* 2 (/ machine-epsilon epsilon)))
         (verbose (send self :verbose)))
    (unless (and (send self :eta) (send self :recycle))
            (send self :initialize-search))
    (send self :compute-step)
    (do ((count 1 (+ count 1))
         (beta 0 (send self :coef-estimates))
         (last-beta -1 beta)
         (dev  min-dev (send self :deviance))
         (last-dev  -1 dev))
        ((or (> count maxcount) 
             (< dev  min-dev)
;            (< (max (abs (/ (- beta last-beta)
;                            (pmax (abs last-beta) low-lim))))
;               epsilon)
             (< (abs (- dev last-dev)) epsilon-dev)))
        (if verbose 
            (format t "Iteration ~d: deviance = ~,6g~%" 
                    count (send self :deviance)))
        (send self :compute-step))))

(defmeth glim-proto :min-deviance ()
 (if (boundp '*deviance-minimum*) *deviance-minimum* .01))

(send glim-proto :epsilon-dev 
 (if (boundp '*deviance-change*) *deviance-change* .001))

;;;
;;; Other fixes for missing data; modified 990122
;;;

(defmeth binomialreg-proto :fit-deviances (mu)
  (flet ((plogp/q (p q) (mapcar #'(lambda (p q)
            (let ((r (if (> q 0) (div1 p q) 0)))  
             (cond ((and (> p 0) (> q 0)) (* p (logpos r)))
                   (t 0)))) p q )))
    (let* ((n (send self :trials))
           (y (send self :yvar))
           (n-y (- n y))
           (n-mu (- n mu))
           (pw (send self :pweights))
           (raw-dev (* 2 (+ (plogp/q y mu) (plogp/q n-y n-mu)))))
      (if pw (* pw raw-dev) raw-dev))))

(defmeth poissonreg-proto :fit-deviances (mu)
  (flet ((plogp/q (p q) (mapcar #'(lambda (p q)
            (let ((r (if (> q 0) (div1 p q) 0)))
             (cond ((and (> p 0) (> q 0)) (* p (logpos r)))
                   (t 0)))) p q )))
    (let* ((y (send self :yvar))
           (raw-dev (* 2 (- (plogp/q y mu) (- y mu))))
           (pw (send self :pweights)))
      (if pw (* pw raw-dev) raw-dev))))

(defmeth gammareg-proto :fit-deviances (mu)
  (let* ((y (send self :yvar))
         (sel (find-obs (+ y mu)))
         (m (select mu sel))
         (y (select y sel))
         (pw (send self :pweights))
         (raw-dev (* 2 (+ (- (log (div y m))) (div (- y m) m))))
         (dev (if pw (* raw-dev (select pw sel)) raw-dev))
         (o (repeat (set-missing) (length mu))))
   (setf (select o sel) dev)
   o))

(defun logpos (x)
 (cond ((eq *computer-type* 'windows)
        (cond ((is-missing x) not-a-number)
               ((< 0 x) (log x))
               (t 0))) 
       (t (log (if-else (< 0 x) x 1)))))

(defmeth glim-proto :compute-step ()
"Args: ()
Executes one iteratively reweighted least squares step."
  (let* ((yvar (send self :yvar))
         (offset (send self :offset))
         (eta (send self :eta))
         (mu (send self :fit-means eta))
         (d-eta (send self :fit-link-derivs mu))
         (z (- (+ eta (* (- yvar mu) d-eta)) offset))
         (v (send self :fit-variances mu))
         (w-inv (* d-eta d-eta v))
         (pw (send self :pweights)))
    (send self :y z)
    (send self :weights (if pw (div pw w-inv) (div w-inv))) ; use of div
    (call-method regression-model-proto :compute)
    (send self :set-eta)
    (send self :set-deviances)
    (if (send self :estimate-scale) 
        (send self :scale (send self :fit-scale)))))

(defmeth glim-proto :fit-means (&optional (eta (send self :eta)))
"Message args: (&optional (eta (send self :eta)))
Retruns mean values for current or supplied ETA."
  (with-missing #'(lambda (v) (send (send self :link) :means v)) eta))

(defmeth glim-proto :initialize-search (&optional eta)
 (let ((arg (send self :initial-means))
       (f #'(lambda (val) (send (send self :link) :eta val))))
  (send self :set-eta (if eta eta (with-missing f arg)))
  (send self :needs-computing t)))

(defmeth glim-proto :fit-link-derivs (mu)
"Message args: ()
Returns link derivative values at MU."
 (let ((arg mu)
       (f #'(lambda (val) (send (send self :link) :derivs val))))
  (with-missing f arg)))

(defmeth normalreg-proto :prediction (&rest args)
 (apply #'call-method regression-model-proto :prediction args))

(defmeth glim-proto :prediction (&rest args)
 (sysbeep)(message-dialog "Prediction implemented only for
multiple linear regression."))
