
;;;$Revision: 1.1 $

;;;
;;; methods for handling models
;;;

(defmeth dataset-proto :delete-model (model)
  (setf (slot-value 'models) (remove model (slot-value 'models))))

(defmeth dataset-proto :models (&optional model)
    (when model (setf (slot-value 'models) 
                      (cons model (slot-value 'models))))
    (slot-value 'models))

(defmeth dataset-proto :model-names ()
  (when (send self :models)
        (mapcar #'(lambda (a) (send a :name)) (slot-value 'models))))

(defmeth dataset-proto :model-type (&optional type)
   (let* ((models (send self :models)))
    (cond ((null models) nil)
          ((null type) models)
          (t (select models (which (mapcar #'(lambda (a)
                  (if (eql type a) t nil)) (send self :model-types))))))))

(defmeth dataset-proto :model-types ()
  (when (send self :models)
        (mapcar #'(lambda (a) (send a :type)) (slot-value 'models))))

(defmeth dataset-proto :find-model (name)
  (when (send self :models)
   (let* ((names (mapcar #'string-upcase (send self :model-names)))
          (p (position (string-upcase name) names :test #'equal)))
      (select (send self :models) p))))

(defmeth dataset-proto :last-model (&key type) 
  (first (send self :model-type type)))

(defmeth dataset-proto :model-number (&optional (update t))
  (when (null (send self :has-slot 'model-number))
        (send self :add-slot 'model-number 0))
  (if update (setf (slot-value 'model-number) (1+ (slot-value 'model-number))))
  (slot-value 'model-number))

;; This next method is needed to allow glms to get weights right be making
;; sure they are defined before any other methods are called.  This needs
;; to be fixed to allow :needs-computing rather than :compute, because this
;; can be terrible if there are lots of models.
(defmeth dataset-proto :update-models ()
; (mapcar #'(lambda (a) (send a :needs-computing t)) (send self :models)))
  (mapcar #'(lambda (a) (send a :compute)) (send self :models)))


;;; regression model methods
;;;

(defmeth regression-model-proto :type (&optional new)
  (if (null (send self :has-slot 'type))
            (send self :add-slot 'type))
  (if new (send self :add-slot 'type new))
  (slot-value 'type))

(defmeth regression-model-proto :remove-menu ()
  (send (send self :menu) :remove)
  (send (send self :data) :delete-model self))

;;;
;;; generalized linear models
;;;

(defmeth dataset-proto :fit-glm (&key (type :binomial) 
       (types *glm-variate-types-default*) (prefix "M") (dialog #'glm-dialog))
 (send self :ones)
 (let* ((name (send self :get-model-name prefix))
        (old-mod (send self :last-model :type type))
        (names (send self :types types))
        (n (if old-mod (reverse (set-difference names 
                (combine (send old-mod :predictors) 
                         (send old-mod :response-name)) :test #'equal)) names))
        (om (if old-mod (send old-mod :args) nil))
        (ans (apply dialog names 
               :name name :candidates n :type type om)))
       (if ans (apply #'send self :make-glm ans))))

(defparameter *show-display-fit* t)

(defmeth dataset-proto :make-glm (&rest args &key predictors response
       intercept weights name offset trials type mean-function)
  (let* ((x (apply #'append (mapcar #'(lambda (n) (send self :data n))
                             predictors)))
         (predictor-names (combine (mapcar #'(lambda (a)
                             (send self :labels a)) predictors)))
         (y (first (send self :data response)))
         (inc (send self :included))
         (response-name response)
         (pweights (send self :get-weights weights))
         (weights-name weights)
         (dataset self)
         (offset-name offset)
         (offset (first (send self :data offset)))
         (trials-name trials)
         (trials (first (send self :data trials)))
         (not-missing (apply #'send self :make-not-missing-indicator 
                    (append (if weights (list pweights))
                            (if trials (list trials-name))
                            (if offset (list offset-name) )
                            (list response) predictors)))
         (link (select (link-functions type)
                (position mean-function (mean-functions type) :test #'equal)))
         r)
     (when (eq type :binomial)
      (let ((nmin (min (select trials (find-obs trials))))
            (ymin (min (select y (find-obs y)))))
       (cond ((< nmin 0) (sysbeep)
              (return-from :make-glm 
               (message-dialog "Number of trials is not always positive.")))
             ((= nmin 0) (sysbeep) 
              (let* ((name (format nil "~a1" trials-name))
                     (str (format nil "~a=(recode-value ~a 0 '?)" 
                                  name trials-name))
                     (fix (ok-or-cancel-dialog (format nil
"Number of trials is zero for some cases.  
Click OK to create a new trials variate
~a with zeroes set to missing.  " name))))
                (when fix
                      (send self :variate
                        :data (recode-value trials 0 '?)
                        :info str
                        :name name)
                       (apply #'send self :make-glm :trials name args))
                (return-from :make-glm)))
             ((< (min (select y (find-obs y))) 0) (sysbeep)
              (return-from :make-glm 
               (message-dialog 
"Number of successes is not always non-negative.")))
             ((> (max (select (- y trials) (find-obs (- y trials)))) 0) 
              (sysbeep)
              (return-from :make-glm 
              (message-dialog 
"Number of successes exceeds number of trials."))))
        ))
     (setf r (send (type-to-proto type link) :new :x x :y y :trials trials
               :predictor-names predictor-names
               :response-name response-name
               :pweights pweights 
               :data self :link link
               :not-missing not-missing
               :included (if not-missing
                    (mapcar #'(lambda (a b) (and a b)) inc not-missing) inc)
               :intercept intercept
               :offset (if offset offset 0)
               :print nil))
     (send self :intern name r)
     (send r :type (if (eq link quad1d-link) :nonlinear type))
     (if (eq link quad1d-link)
         (defmeth r :plotlist ()
          (append (call-next-method) (list :linear-part))))
     (send r :add-slot 'args args)
     (send r :delete-slot 'included)
     (defmeth r :included (&rest args) 
      (let* ((inc (apply #'send dataset :included args))
             (obs (send self :not-missing)))
       (if obs (mapcar #'(lambda (a b) (and a b)) inc obs) inc)))
     (when *show-display-fit* (send r :display))
     (unless (send r :needs-computing) 
             (send r :menu name)
             (send self :models r))
     r))

(defmeth dataset-proto :get-weights (w)
  (let* ((vals (first (send self :data w)))
         (ok (cond (vals (> (min vals) 0)) (t t))))
   (cond (ok vals)
         (t (sysbeep) (message-dialog (format nil
"Some cases have weights that are zero or negative.
These cases are set to missing in fitting this model."))
            (truncate-at-val vals 0 :value (set-missing))))))

(defmeth dataset-proto :make-not-missing-indicator (&rest args)
  (let* ((m (apply #'append (mapcar #'(lambda (a) (send a :missing))
             (mapcar #'(lambda (d) (send self :find-datalist d)) args))))
         ans)
  (cond ((null m) nil)
        (t (setf ans (make-list (send self :num-cases) :initial-element t))
           (setf (select ans m) nil)
           ans))))

(defun glm-dialog (names &key 
          (type :binomial)
          (name "reg") (offset "")
          (weights "") (trials "")
          (response "") mean-function
          (candidates names)
          (predictors (repeat "" (length names)))
          (intercept t))
"Function args: (names &key reg-name type)
Opens a dialog box to define a regression."
  (let* ((version (send text-item-proto :new
               (format nil "~a" *arc-name* )))
         (all (send allocate-items-proto :new))
         (candidates-l (send subordinate-list-proto :new 
                             (pad-list candidates (length names))
                             'primary all))
         (candidates-n (send text-item-proto :new "Candidates"))
         (predictors-l (send subordinate-list-proto :new 
                             (pad-list predictors (length names))
                             'secondary all))
         (predictors-n (send text-item-proto :new "Terms/Predictors"))
         (response-l (send subordinate-list-proto :new (list response)
                             'oneitem all))
         (response-n (send text-item-proto :new "Response...    "))
         (weight-l (send subordinate-list-proto :new 
                   (if weights (list weights) '("")) 'oneitem all))
         (weight-n (send text-item-proto :new "Weights...     "))
         (offset-l (send subordinate-list-proto :new
                   (if offset (list offset) '("")) 'oneitem all))
         (offset-n (send text-item-proto :new "Offset...      "))
         (trials-l (when (eql type :binomial)
                   (list (send subordinate-list-proto :new
                   (if trials (list trials) '("")) 'oneitem all))))
         (trials-n (when (eql type :binomial)
                   (list (send text-item-proto :new "Trials...      "))))
         (intercept-n (send toggle-item-proto :new "Fit Intercept" 
                            :value intercept))
         (mod-type (format nil "~a Regression" (string-capitalize type)))
         (name-l (send edit-text-item-proto :new name :text-length 
                       (apply #'ms *text-width-short*) ))
         (name-n (send text-item-proto :new 
                     (format nil "Name for ~%~a" mod-type)))
         (mfs (mean-functions type))
         (mfs-l (send choice-item-proto :new mfs :value 
                   (position mean-function mfs :test #'equal)))
         (mfs-n (send text-item-proto :new "Kernel Mean Function"))
         (get-answer #'(lambda ()
           (let* ((ans (send all :value))
                  (raw (send all :all-values)))
             (cond ((or (null (select ans 1)) (null (select ans 2))
                    (when trials-n (null (select ans 5))))
                    (message-dialog "Predictor, response or trials missing")
                    (glm-dialog names :type type
                              :name (send name-l :text)
                              :candidates (select raw 0)
                              :predictors (select raw 1)
                              :response (first (select raw 2))
                              :weights (first (select raw 3))
                              :offset (first (select raw 4))
                              :trials (when trials-n (first (select raw 5)))
                              :intercept (send intercept-n :value)))
                   (t (list :predictors (select ans 1)
                            :candidates (select ans 0)
                            :type type
                            :response (first (select ans 2))
                            :weights (first (select ans 3))
                            :offset (first (select ans 4))
                            :trials (when trials-n (first (select ans 5)))
                            :mean-function (select mfs (send mfs-l :value))
                            :intercept (send intercept-n :value)
                            :name (send name-l :text)))))))
         (cancel (send modal-button-proto :new "Cancel"))
         (done (send modal-button-proto :new "OK" :action 
                     #'(lambda () (funcall get-answer))))
         (dialog (send modal-dialog-proto :new
                       (list (list version name-n name-l)
                             (list (append (list candidates-n candidates-l 
                                          response-n weight-n
                                         offset-n) trials-n)
                                    (append (list predictors-n predictors-l 
                                         response-l weight-l
                                         offset-l) trials-l)
                                   (list intercept-n 
                                         mfs-n mfs-l done cancel))))))
        (send dialog :default-button done)
        (send dialog :modal-dialog) ))

(defun multinomial-dialog (names &key 
          (type :binomial)
          (name "reg") (offset "")
          (weights "") (trials "")
          (response "") mean-function
          (candidates names)
          (predictors (repeat "" (length names)))
          (intercept t))
"Function args: (names &key reg-name type)
Opens a dialog box to define a regression."
  (let* ((version (send text-item-proto :new
               (format nil "~a" *arc-name* )))
         (all (send allocate-items-proto :new))
         (candidates-l (send subordinate-list-proto :new 
                             (pad-list candidates (length names))
                             'primary all))
         (candidates-n (send text-item-proto :new "Candidates"))
         (predictors-l (send subordinate-list-proto :new 
                             (pad-list predictors (length names))
                             'secondary all))
         (predictors-n (send text-item-proto :new "Terms/Predictors"))
         (response-l (send subordinate-list-proto :new (list response)
                             'oneitem all))
         (response-n (send text-item-proto :new "Response...    "))
         (weight-l (send subordinate-list-proto :new 
                   (if weights (list weights) '("")) 'oneitem all))
;         (weight-n (send text-item-proto :new "Weights...     "))
         (weight-n (send text-item-proto :new "Frequencies...     "))

         (offset-l (send subordinate-list-proto :new
                   (if offset (list offset) '("")) 'oneitem all))
         (offset-n (send text-item-proto :new "Offset...      "))
         (trials-l (when (eql type :binomial)
                   (list (send subordinate-list-proto :new
                   (if trials (list trials) '("")) 'oneitem all))))
         (trials-n (when (eql type :binomial)
                   (list (send text-item-proto :new "Trials...      "))))
;         (intercept-n (send toggle-item-proto :new "Fit Intercept" 
;                            :value intercept))
         (mod-type (format nil "~a Regression" (string-capitalize type)))
         (name-l (send edit-text-item-proto :new name :text-length 
                       (apply #'ms *text-width-short*) ))
         (name-n (send text-item-proto :new 
                     (format nil "Name for ~%~a" mod-type)))
         (mfs (mean-functions type))
         (mfs-l (send choice-item-proto :new mfs :value 
                   (position mean-function mfs :test #'equal)))
         (mfs-n (send text-item-proto :new "Kernel Mean Function"))
         (get-answer #'(lambda ()
           (let* ((ans (send all :value))
                  (raw (send all :all-values)))
             (cond ((or (null (select ans 1)) (null (select ans 2))
                    (when trials-n (null (select ans 5))))
                    (message-dialog "Predictor, response or trials missing")
                    (glm-dialog names :type type
                              :name (send name-l :text)
                              :candidates (select raw 0)
                              :predictors (select raw 1)
                              :response (first (select raw 2))
                              :weights (first (select raw 3))
                              :offset (first (select raw 4))
                              :trials (when trials-n (first (select raw 5)))
 ;                             :intercept (send intercept-n :value)
                              ))
                   (t (list :predictors (select ans 1)
                            :candidates (select ans 0)
                            :type type
                            :response (first (select ans 2))
                            :weights (first (select ans 3))
                            :offset (first (select ans 4))
                            :trials (when trials-n (first (select ans 5)))
                            :mean-function (select mfs (send mfs-l :value))
 ;                           :intercept (send intercept-n :value)
                            :name (send name-l :text)))))))
         (cancel (send modal-button-proto :new "Cancel"))
         (done (send modal-button-proto :new "OK" :action 
                     #'(lambda () (funcall get-answer))))
         (dialog (send modal-dialog-proto :new
                       (list (list version name-n name-l)
                             (list (append (list candidates-n candidates-l 
                                          response-n weight-n
                                         offset-n) trials-n)
                                    (append (list predictors-n predictors-l 
                                         response-l weight-l
                                         offset-l) trials-l)
                                   (list ;intercept-n 
                                         mfs-n mfs-l done cancel))))))
        (send dialog :default-button done)
        (send dialog :modal-dialog) ))

